import copy
from dataclasses import dataclass
import dataclasses
from typing import Dict, Tuple, Union

import gym
from gym import error, spaces, utils
from gym.utils import seeding

from wumpus import WumpusWorld, Hunter


class WumpusEnv(gym.Env):
    metadata = {'render.modes': ['human', 'ansi']}
    actions = tuple(Hunter.Actions)

    def __init__(self, world: WumpusWorld = None):
        """Initialise a new Gym environment with an instance of the Wumpus world
        
        Keyword Arguments:
            world {WumpusWorld} -- Wumpus world, if None a random one will be created (default: {None})
        """
        self.__initial_world: WumpusWorld = world if isinstance(world, WumpusWorld) else WumpusWorld.classic()
        # make sure there's an hunter
        try:
            next(iter(o for o in self.__initial_world.objects if isinstance(o, Hunter)))
        except StopIteration:
            raise ValueError('Missing hunter in the Wumpus environment {}'.format(self.__initial_world))
        self.__world: WumpusWorld = None
        self.__agent: Hunter = None
        self.__done: bool = False

        # Actions are discrete integer values
        self.action_space = spaces.Discrete(len(self.actions))

        self.observation_space = spaces.Dict({f.name: spaces.Discrete(2) for f in dataclasses.fields(Hunter.Percept)})

    def _percept_to_space(self) -> Dict[str, int]:
        percept = self.__agent.percept()
        return {k: (1 if v else 0) for k, v in dataclasses.asdict(percept).items()}

    @classmethod
    def space_to_percept(cls, obs: Dict[str, int]) -> Hunter.Percept:
        """Converts the Gym ` <http://gym.openai.com/docs/#spaces>` to the Wumpus environment percept
        
        Arguments:
            obs {Dict[str, int]} -- an observation space for the environment, see ``WumpusEnv.observation_space`` property for details
        
        Returns:
            Hunter.Percept -- named tuple corresponding to the observation
        """
        return Hunter.Percept(**{k: v > 0 for k, v in obs.items()})

    @classmethod
    def space_to_action(cls, action: int) -> Hunter.Actions:
        """Converts the integer in the environment action space to the corresponding action in the ``Hunter`` agent
        
        Arguments:
            action {int} -- an integer from zero to the number of actions minus one, see ``WumpusEnv.action_space`` property for details
        
        Returns:
            Hunter.Actions -- the corresponding action
        """
        return cls.actions[action]

    def step(self, action: Union[int, Hunter.Actions]) -> Tuple[Dict[str, int], float, bool, Dict]:
        """OpenAI Gym ``step`` method, see `Gym documentation <http://gym.openai.com/docs/#observations>` for details
        
        Arguments:
            action {Union[int, Hunter.Actions]} -- the method accepts both an action space value (integer) or a ``Hunter.Actions`` object
        
        Returns:
            Tuple[Dict[str, int], float, bool, Dict] -- see ``gym.step`` method for details
        """
        if self.__world is None:
            self.reset()
        if self.__done:
            # episode ended, undefined result
            return (None, 0.0, True, {})
        info = {'alive': True}
        reward = self.__agent.do(self.actions[action] if isinstance(action, int) else action)
        if self.__agent.success():
            self.__done = True
            info['success'] = True
        if not self.__agent.isAlive:
            self.__done = True
            info['alive'] = False
        return (self._percept_to_space(), float(reward), self.__done, info)

    def reset(self) -> Hunter.Percept:
        self.__world = copy.deepcopy(self.__initial_world)
        self.__agent = next(iter(o for o in self.__world.objects if isinstance(o, Hunter)), None)
        self.__done = False
        return self._percept_to_space()

    def render(self, mode='human'):
        if self.__world is None:
            self.reset()
        if mode == 'human':
            print(self.__world)
        elif mode == 'ansi':
            return str(self.__world)
        else:
            super().render(mode=mode)

    def close(self):
        self.__world = None
        self.__agent = None
        self.__done = False

    @classmethod
    def from_dict(cls, desc: Dict):
        """Creates a new WumpusEnv object from a dictionary object with the world description.
        
        Arguments:
            desc {Dict} -- the description of the Wumpus World, see ``WumpusWorld.from_JSON`` method for details
        
        Returns:
            WumpusEnv -- a new object
        """
        world = WumpusWorld.from_JSON(desc)
        return cls(world=world)

    @classmethod
    def classic(cls, size: int = 4, seed=None):
        """Creates a new WumpusEnv object with pits and Wumpus positions randomly generated.
        
        Keyword Arguments:
            size {int} -- the size of the grid (default: {4})
            seed {int} -- seed to initialise the random generator (default: {None})
        
        Returns:
            WumpusEnv -- a new object
        """
        world = WumpusWorld.classic(size=size, seed=seed)
        return cls(world=world)


def wumpusenv_from_dict(desc: Dict) -> WumpusEnv:
    """Creates a new WumpusEnv object from a dictionary object with the world description.
    
    Arguments:
        desc {Dict} -- the description of the Wumpus World, see ``WumpusWorld.from_JSON`` method for details
    
    Returns:
        WumpusEnv -- a new object
    """
    return WumpusEnv.from_dict(desc=desc)


def wumpusenv_classic(size: int = 4, seed=None) -> WumpusEnv:
    """Creates a new WumpusEnv object with pits and Wumpus positions randomly generated.
    
    Keyword Arguments:
        size {int} -- the size of the grid (default: {4})
        seed {int} -- seed to initialise the random generator (default: {None})
    
    Returns:
        WumpusEnv -- a new object
    """
    return WumpusEnv.classic(size=size, seed=seed)
