import random
import textwrap
from typing import Iterable

from .gridworld import Agent, Percept, object_id, GridWorld


class Player(object):
    """A player for a given agent. It implements the play method which should
    return one of the actions for the agent or None to give up.
    """
    def __init__(self, name: str = None):
        """
        Initialise the name of the agent if provided.
        """
        if name is not None:
            self.name = str(name)

    def start_episode(self):
        """Method called at the beginning of the episode."""
        pass

    def end_episode(self, outcome: int, alive: bool, success: bool):
        """Method called at the when an episode is completed with the outcome of the game and whether the agent was still alive and successfull.
        """
        pass

    def play(self, turn: int, percept: Percept, actions: Iterable[Agent.Actions]) -> Agent.Actions:
        """Given a turn (integer) and a percept, which might differ according to the specific problem, returns an action, among the given list of possible actions, to play at the given turn or None to stop the episode."""
        raise NotImplementedError

    def feedback(self, action: Agent.Actions, reward: int, percept: Percept):
        """Receive in input the reward of the last action and the resulting state. The function is called right after the execution of the action."""
        pass

    @property
    def name(self) -> str:
        """Return the name of the player or a default value based on its type and hash."""
        try:
            return self._name
        except AttributeError:
            return object_id(self)

    @name.setter
    def name(self, value: str):
        """Set the name of the player.

        Args:
            value (str): the name
        """
        self._name = value


class InformedPlayer(Player):
    """
    A player that receives the configuration of the world at the beginning of the episode.
    """
    def start_episode(self, word: GridWorld):
        """Method called at the beginning of the episode. Receives the current world status."""
        pass


class UninformedPlayer(Player):
    """
    A player that doesn't know what is the configuration of the environment.
    """

    def start_episode(self):
        """Method called at the beginning of the episode. Receives the current world status."""
        pass


############################################################################
#
#  Examples of the use of the API


#######################################
#   Trivial players
#


class RandomPlayer(Player):
    """This player selects randomly one of the available actions."""
    def play(self, turn: int, percept: Percept, actions: Iterable[Agent.Actions]) -> Agent.Actions:
        actions_lst = list(actions)
        return actions_lst[random.randint(0, len(actions) - 1)]


class UserPlayer(Player):
    """This player asks the user for the next move, if it's not ambiguous it accepts also commands initials and ignores the case."""
    def play(self, turn: int, percept: Percept, actions: Iterable[Agent.Actions]) -> Agent.Actions:
        actions_dict = {a.name: a for a in actions}
        print('{} percept:'.format(self.name))
        print(textwrap.indent(str(percept), '  '))
        while True:
            answer = input('{}: select an action {} and press enter, or empty to stop: '.format(self.name, list(actions_dict.keys()))).strip()
            if len(answer) < 1:
                return None
            elif answer in actions_dict:
                return actions_dict[answer]
            else:
                options = [k for k in actions_dict.keys() if k.lower().startswith(answer.lower())]
                if len(options) == 1:
                    return actions_dict[options[0]]
                else:
                    print('Cannot understand <{}>'.format(answer))


