#!/usr/bin/env python

"""
Functions to run the players
"""

import argparse
import copy
import importlib
import io
import json
import os
import re
import sys
from typing import Any, Dict, Iterator, Type

from .gridworld import Agent, GridWorld, GridWorldException, EaterWorld
from .player import Player, InformedPlayer
from .wumpus import WumpusWorld


def get_subclasses(cls):
    for subclass in cls.__subclasses__():
        yield from get_subclasses(subclass)
        yield subclass


def check_entrypoint(arg_value: str, pattern: re.Pattern = re.compile(r"^[\w.-]+:[\w.-]+$")) -> str:
    """Checks that the argument is a valid object reference specification in the form 'importable.module:object.attr'. See <https://packaging.python.org/specifications/entry-points/#data-model> for details.

    Args:
        arg_value (str): object reference
        pattern ([type], optional): regular expression defining the required reference. Defaults to '^[\w.-]+:[\w.-]+$'.

    Raises:
        argparse.ArgumentTypeError: if the argument doesn't comply with the pattern

    Returns:
        str: the given argument if it complies with the pattern
    """
    if not pattern.match(arg_value):
        raise argparse.ArgumentTypeError("the argument doesn't comply with {}".format(pattern.pattern))
    return arg_value


def get_player_class(object_ref: str, path: os.PathLike = None) -> Type[Player]:
    if path is not None and path not in sys.path:
        if not os.path.isdir(path):
            raise FileNotFoundError('Directory <{}> not found'.format(path))
        sys.path.insert(0, path)

    # see <https://packaging.python.org/specifications/entry-points/#data-model>
    modname, qualname_separator, qualname = object_ref.partition(':')
    obj = importlib.import_module(modname)
    if qualname_separator:
        for attr in qualname.split('.'):
            try:
                obj = getattr(obj, attr)
            except AttributeError as e:
                raise ImportError('Cannot import {} object: {}'.format(object_ref, e))

    player_class = obj
    if not issubclass(player_class, Player):
        raise NotImplementedError('class {} is not a subclass of Player'.format(player_class))
    return player_class


def get_world_class(name: str) -> Type[GridWorld]:
    """Return the class of the world corresponding to the given name.

    Args:
        name (str): name of the GridWorld subclass

    Returns:
        Type[GridWorld]: GridWorld subclass
    """
    world_class = globals().get(name, None)
    if world_class is None:
        raise NotImplementedError('GridWorld subclass {} is not available'.format(name))
    if not issubclass(world_class, GridWorld):
        raise NotImplementedError('Class {} is not a subclass of GridWorld'.format(name))

    return world_class


def worlds(files, world_class: Type[GridWorld]) -> Iterator[GridWorld]:
    for fd in files:
        try:
            world_defs = json.load(fd)
            if isinstance(world_defs, dict):
                yield world_class.from_dict(world_defs)
            elif isinstance(world_defs, list):
                for wd in world_defs:
                    try:
                        yield world_class.from_dict(wd)
                    except Exception as e:
                        print('Skipping world {}: {}'.format(wd, e))
                        continue
        except Exception as e:
            print('Skipping {}: {}'.format(fd, e))
            continue


def run_episode(world: GridWorld, player: Player, agent: Agent = None, horizon: int = 0, show=True, outf: io.TextIOBase = None) -> Dict[str, Any]:
    """Run an episode on the world using the player to control the agent. The horizon specifies the maximum number of steps, 0 or None means no limit. If show is true then the world is printed ad each iteration before the player's turn.

        Raise the exception GridWorldException is the agent is not in the world.

    Args:
        world (GridWorld): the world in which the episode is run
        player (Player): the player
        agent (Agent, optional): the agent controlled by the player. Defaults to first agent in the world
        horizon (int, optional): stop after this number of steps, 0 for no limit. Defaults to 0.
        show (bool, optional): whether to show the environment before a step. Defaults to True.
        outf (TextIOBase, optional): writes output to the given stream. Defaults to stdout.

    Returns:
        dictionary (JSON encodable) with the log of the game

    Raises:
        GridWorldException: if there are problems with the world (e.g. there's no agent)
    """
    if outf is None:
        outf = sys.stdout
    if agent is None:
        try:
            agent = next(o for o in world.objects if isinstance(o, Agent))
        except StopIteration:
            raise GridWorldException(f'No agent in this {world}')
    elif agent not in world.objects:
        raise GridWorldException('Missing agent {}, cannot run the episode'.format(agent))
    game_log = {
        'world': world.to_dict(),
        'agent': agent.name,
        'player': player.name,
        'actions': [],
        'exceptions': [],
        'maxsteps': False
    }
    # inform the player of the start of the episode
    try:
        if isinstance(player, InformedPlayer):
            player.start_episode(copy.deepcopy(world))
        else:
            player.start_episode()
    except Exception as e:
        print(f'Exception in Player.start_episode: {e}', file=outf)
        game_log['exceptions'].append(f'Player.start_episode: {e}')
    step = 0
    while not horizon or step < horizon:
        if agent.success():
            print('The agent {} succeeded!'.format(agent.name), file=outf)
            break
        if not agent.isAlive:
            print('The agent {} died!'.format(agent.name), file=outf)
            break

        if show:
            print(world, file=outf)
        try:
            action = player.play(step, agent.percept(), agent.actions())
        except Exception as e:
            action = None
            print(f'Exception in Player.play: {e}', file=outf)
            game_log['exceptions'].append(f'Player.play: {e}')
        if action is None:
            game_log['actions'].append(action)
            agent.suicide()
            print('Episode terminated by the player {}.'.format(player.name), file=outf)
            break
        game_log['actions'].append(action.name)
        reward = agent.do(action)
        print('Step {}: agent {} executing {} -> reward {}'.format(step, agent.name, action.name, reward), file=outf)
        try:
            player.feedback(action, reward, agent.percept())
        except Exception as e:
            print(f'Exception in Player.feedback: {e}', file=outf)
            game_log['exceptions'].append(f'Player.feedback: {e}')

        step += 1
    else:
        print('Episode terminated by maximum number of steps ({}).'.format(horizon), file=outf)
        game_log['maxsteps'] = True

    try:
        player.end_episode(agent.reward, agent.isAlive, agent.success())
    except Exception as e:
        print(f'Exception in Player.end_episode: {e}', file=outf)
        game_log['exceptions'].append(f'Player.end_episode: {e}')

    game_log['reward'] = agent.reward
    game_log['alive'] = agent.isAlive
    game_log['success'] = agent.success()

    print(world, file=outf)
    print('Episode terminated with a reward of {} for agent {}'.format(agent.reward, agent.name), file=outf)

    return game_log
